#!/usr/bin/env python
# CompareTextOutputDef.py - contains defs for comparing textual output from eddy
# It has code for comparing two .eddy_movement_over_time files and for comparing
# two .eddy_outlier_n_stdev_map files.
#

from __future__ import print_function # I have no words...

import            collections
import            datetime
import            tempfile
import            logging
import            shutil
import            fnmatch
import            re
import            sys
import            os
import os.path as op
import numpy   as np

def MovementOverTimeAreSame(testfname, reffname, maxtrans, trans99, meantrans, maxrot, rot99, meanrot, skip, *args):
    """Compares two .eddy_movement_over_time files and returns true or false.
    Pass or fail depends on observed differences compared to limits specified by
    inputs maxtrans, trans99, meantrans, maxrot, rot99 and meanrot. Skip is 0 or 1 if 
    we are to skip comparison for x-trans (0) or y-trans (1). If not, skip should be -1.
    USE: MovementOverTimeAreSame(testfname, reffname, maxtrans, trans99, meantrans, maxrot, rot99, meanrot, skip)
    or
    MovementOverTimeAreSame(testfname, reffname, maxtrans, trans99, meantrans, maxrot, rot99, meanrot, skip, nsl, fsl, lsl)
    """

    if (skip < -1) or (skip > 1):
        print('MovementOverTimeAreSame: Invalid skip value',file=sys.stderr)
        raise ValueError('MovementOverTimeAreSame: Invalid skip value')

    if len(args) != 0 and len(args) != 3:
        print('MovementOverTimeAreSame: Invalid number of optional arguments',file=sys.stderr)
        raise RuntimeError('MovementOverTimeAreSame: Invalid number of optional arguments')

    if len(args) == 3:
        nsl = args[0]
        fsl = args[1]
        lsl = args[2]
        if fsl < 0 or lsl > (nsl-1):
            print('MovementOverTimeAreSame: Invalid fsl or lsl',file=sys.stderr)
            raise ValueError('MovementOverTimeAreSame: Invalid fsl or lsl')

    try:
        test_data = np.genfromtxt(testfname)
        ref_data = np.genfromtxt(reffname)
    except Exception as e:
        print('MovementOverTimeAreSame: File read error: {}'.format(e.args[0]),file=sys.stderr)
        raise e

    if test_data.shape != ref_data.shape or test_data.shape[1] != 6 or test_data.shape[0] % nsl != 0:
        print('MovementOverTimeAreSame: Shape mismatch between files {} and {}'.format(testfname,reffname),file=sys.stderr)
        raise ValueError('MovementOverTimeAreSame: Shape mismatch between files {} and {}'.format(testfname,reffname))

    # Remove end slices if nsl, fsl and lsl have been set
    if len(args) == 3:
        try:
            if test_data.shape[0] % nsl:
                raise ValueError('')
            for p in range(0,6):
                tmp = test_data[:,p]
                tmp = np.reshape(tmp,(nsl,test_data.shape[0]//nsl),'F') 
                tmp = tmp[fsl:lsl+1,:]
                if p == 0:
                    tmp_test_data = np.reshape(tmp,(tmp.size,1),'F')
                else:
                    tmp_test_data = np.concatenate((tmp_test_data,np.reshape(tmp,(tmp.size,1),'F')),axis=1)
                tmp = ref_data[:,p]
                tmp = np.reshape(tmp,(nsl,test_data.shape[0]//nsl),'F') 
                tmp = tmp[fsl:lsl+1,:]
                if p == 0:
                    tmp_ref_data = np.reshape(tmp,(tmp.size,1),'F')
                else:
                    tmp_ref_data = np.concatenate((tmp_ref_data,np.reshape(tmp,(tmp.size,1),'F')),axis=1)
        except Exception as e:
            print('MovementOverTimeAreSame: Input nsl={}, fsl={} or lsl={} not suited for files {} and {}'.format(nsl,fsl,lsl,testfname,reffname),file=sys.stderr)
            raise ValueError('MovementOverTimeAreSame: Input nsl={}, fsl={} or lsl={} not suited for files {} and {}'.format(nsl,fsl,lsl,testfname,reffname))



    test_data = tmp_test_data
    ref_data = tmp_ref_data


    diff = test_data - ref_data
    absdiff = np.absolute(diff)
    if skip == 0:
        atdiff = absdiff[:,[1, 2]]
    elif skip == 1:
        atdiff = absdiff[:,[0, 2]]
    elif skip == -1:
        atdiff = absdiff[:,[0, 1, 2]]
    
    ardiff = 180.0 * absdiff[:,[3, 4, 5]] / 3.141593    
    
    omaxtrans = np.amax(atdiff)                                    # Max translation
    omaxrot = np.amax(ardiff)                                      # Max rotation
    omeantrans = np.max(np.mean(atdiff,0))                         # Mean translation along worst direction
    omeanrot = np.max(np.mean(ardiff,0))                           # Mean rotation around worst axis
    otrans99 = np.max(np.sort(atdiff,0)[int(0.99*len(atdiff)),:])  # 99th percentile along worst direction
    orot99 = np.max(np.sort(ardiff,0)[int(0.99*len(atdiff)),:])    # 99th percentile around worst axis

    print('MovementOverTimeAreSame: maxtrans = {}, maxrot = {}'.format(maxtrans,maxrot),file=sys.stdout)
    print('MovementOverTimeAreSame: omaxtrans = {}, omaxrot = {}'.format(omaxtrans,omaxrot),file=sys.stdout)
    print('MovementOverTimeAreSame: meantrans = {}, meanrot = {}'.format(meantrans,meanrot),file=sys.stdout)
    print('MovementOverTimeAreSame: omeantrans = {}, omeanrot = {}'.format(omeantrans,omeanrot),file=sys.stdout)
    print('MovementOverTimeAreSame: trans99 = {}, rot99 = {}'.format(trans99,rot99),file=sys.stdout)
    print('MovementOverTimeAreSame: otrans99 = {}, orot99 = {}'.format(otrans99,orot99),file=sys.stdout)

    rval = True
    if (omaxtrans > maxtrans) or (omaxrot > maxrot) or (otrans99 > trans99) or (orot99 > rot99) or (omeantrans > meantrans) or (omeanrot > meanrot):
        rval = False
    
    return rval

def OutlierNStdevMapsAreTheSame(testfname, reffname, maxdiff1, meandiff1, maxdiff2, meandiff2, *args):
    """Compares two .eddy_outlier_n_stdev_map files and returns true or false.
    Pass or fail depends on observed differences compared to limits specified by
    inputs maxdiff1, meandiff1, maxdiff2 and meandiff2.
    USE: OutlierNStdevMapsAreTheSame(testfname, reffname, maxdiff1, meandiff1, maxdiff2, meandiff2)
    or
    OutlierNStdevMapsAreTheSame(testfname, reffname, maxdiff1, meandiff1, maxdiff2, meandiff2, nsl, fsl, lsl)
    """

    if len(args) != 0 and len(args) != 3:
        print('OutlierNStdevMapsAreTheSame: Invalid number of optional arguments',file=sys.stderr)
        raise ValueError('OutlierNStdevMapsAreTheSame: Invalid number of optional arguments')
    if len(args) == 3:
        nsl = args[0]
        fsl = args[1]
        lsl = args[2]
        if fsl < 0 or lsl > (nsl-1):
            print('OutlierNStdevMapsAreTheSame: Invalid fsl or lsl',file=sys.stderr)
            raise ValueError('OutlierNStdevMapsAreTheSame: Invalid fsl or lsl')

    try:
        test_data = np.genfromtxt(testfname,skip_header=1)
        ref_data = np.genfromtxt(reffname,skip_header=1)
    except Exception as e:
        print('OutlierNStdevMapsAreTheSame: File read error: {}'.format(e.args[0]),file=sys.stderr)
        raise e

    if test_data.shape != ref_data.shape:
        print('OutlierNStdevMapsAreTheSame: Shape mismatch between files {} and {}'.format(testfname,reffname),file=sys.stderr)
        raise ValueError('OutlierNStdevMapsAreTheSame: Shape mismatch between files {} and {}'.format(testfname,reffname))

    if len(args) == 3:
        if nsl != ref_data.shape[1]:
            print('OutlierNStdevMapsAreTheSame: Shape mismatch between the files {} and {} and nsl={}'.format(testfname,reffname,nsl),file=sys.stderr)
            raise ValueError('OutlierNStdevMapsAreTheSame: Shape mismatch between the files {} and {} and nsl={}'.format(testfname,reffname,nsl))
        ref_data = ref_data[:,fsl:lsl+1]
        test_data = test_data[:,fsl:lsl+1]

    try:
        # Find vol-slice tuples that are non-zero in both runs
        nztest = test_data != 0
        nzref = ref_data != 0
        if float(np.sum(nzref)-np.sum(nztest)) / np.sum(nzref) > 0.01: # Ensure ~ equal no. of non-zero
            print('OutlierNStdevMapsAreTheSame: Large mismatch in number of slices considered',file=sys.stdout)
            return False

        mask = np.multiply(nztest.astype('int'),nzref.astype('int'))
        indx = np.where(mask > 0)
        test_data = test_data[indx]
        ref_data = ref_data[indx]

        # Calculate difference-ratio for all non-zero tuples
        radiff = np.divide(np.absolute(test_data - ref_data),np.absolute(ref_data))

        # Check for max and mean diff-ratio for voxels with z > 1
        indx2 = np.where(ref_data > 1.0)
        omaxdiff1 = np.max(radiff[indx2])
        omeandiff1 = np.mean(radiff[indx2])

        # Check for max and mean diff-ratio for voxels with z > 2    
        indx2 = np.where(ref_data > 2.0)
        omaxdiff2 = np.max(radiff[indx2])
        omeandiff2 = np.mean(radiff[indx2])

        print('OutlierNStdevMapsAreTheSame: maxdiff1 = {}, omaxdiff1 = {}'.format(maxdiff1,omaxdiff1),file=sys.stdout)
        print('OutlierNStdevMapsAreTheSame: meandiff1 = {}, omeandiff1 = {}'.format(meandiff1,omeandiff1),file=sys.stdout)
        print('OutlierNStdevMapsAreTheSame: maxdiff2 = {}, omaxdiff2 = {}'.format(maxdiff2,omaxdiff2),file=sys.stdout)
        print('OutlierNStdevMapsAreTheSame: meandiff2 = {}, omeandiff2 = {}'.format(meandiff2,omeandiff2),file=sys.stdout)

    except Exception as e:
        print('OutlierNStdevMapsAreTheSame: Unknown error: {}'.format(e.args[0]),file=sys.stderr)
        raise e


    rval = True
    if (omaxdiff1 > maxdiff1) or (omeandiff1 > meandiff1) or (omaxdiff2 > maxdiff2) or (omeandiff2 > meandiff2):
        rval = False

    return rval
